#!/bin/bash

shellfu import exit
shellfu import inigrep
shellfu import pretty
shellfu import saturnin_help

#
# Saturnin - Spirit of your toolbox
#
# Saturnin is a Shellfu module that can help you build consistent and easy
# to use command line-based toolkit.  The idea is that you have a set of
# scripts that together form a particular tool, which you then integrate
# under a single meta-command (similar to git, apt, systemctl and many
# others...).
#
# Saturnin's main goal is to help you with the integration so that you can
# fully concentrate on development of the scripts, while Saturnin will
# provide the common part of command-line interface.  Adding to that,
# Saturnin also provides several extra features such as config management,
# help system, hooks or support for command line completion.
#
# It's possible to integrate any kind of programs within Saturnin, although
# extra features are currently only available for Shellfu/Bash scripts.
#
# Starting your toolkit
# =====================
#
# The starting point of a Saturnin-powered toolkit is the main meta-command,
# which calls calls saturnin__main(), defining all app-specific parameters.
# You could implement that one yourself, but it's **strongly recommended**
# to start off by creating a copy of 'app.skel' from 'saturnin-demo':
#
#     https://gitlab.com/vornet/infra/saturnin-demo/
#
# In fact, it's best to actually copy the *whole* project, which serves
# as skeleton of not just the meta-command (by Saturnin) but also MKit
# install/build system, which provides features like installation
# scripts, SemVer-compatible versioning and packaging templates for both
# RPM and DEB:
#
#     git clone https://gitlab.com/vornet/infra/saturnin-demo
#     mv saturnin-demo myproject
#     cd myproject
#     rm -rf .git
#     git init
#
# You will need to edit few files, for starters:
#
#     edit mkit.ini                     # your main project info
#     edit packaging/template.spec      # file lists, Requires and description
#     edit packaging/debian/control     # ..also for Debian
#     git add .
#     git commit -m "Initial commit"
#
# ...but the reward is sweet:
#
#     make install      # to install to current system
#
#     make rpmstuff     # to get .spec file and source tarball
#     make debstuff     # to get debian directory
#     make vbump        # to seal off release
#     make release      # to seal off release
#

saturnin__bug() {
    #
    # Warn about bug in your software
    #
    # Issue warning using warn() from pretty but also add application
    # version.  This is useful when an assertion in your application fails
    # such that it is certain that there is a bug inside it.
    #
    # In such cases, it makes sense to print also version information to
    # help users with reporting.
    #
    local msg           # message core
    for msg in "$@"; do
        warn "bug: $msg"
    done
    warn "bug in $(basename "$0") version: $_SATURNIN_APP_VERSION"
}

saturnin__conf() {
    #
    # inigrep smart loader
    #
    # Usage:
    #   saturnin__conf [-j] [INIGREP_QUERY] [-- [FNAME]..]
    #
    #
    # File arguments
    # ==============
    #
    # Each FNAME argument is then processed as follows:
    #
    #  *  `-` (single dash) is interpreted as reading from standard input.
    #
    #  *  If argument contains slash, it is expanded as a regular path
    #     (relative or absolute).
    #
    #  *  Otherwise, it is taken as filename and searched for in directories
    #     given in `$_SATURNIN_CONF_PATH`.  (This can yield more than one
    #     path, which is equivalent as if all paths were provided.)
    #
    #     Not all files expanded based on `$_SATURNIN_CONF_PATH` are read by
    #     default; reading is governed by "merge strategy": the default
    #     strategy "first" reads only the first existing file.
    #
    #     "join" strategy on the other hand, means that any files are simply
    #     concatenated and prefixed with comment (visible only in raw mode)
    #     containing path to the file.
    #
    #     This means that if a section is queried that is present in both
    #     files, it is effectively concatenated as well.
    #
    # If no FNAME argument is passed, filename is inferred fron INIGREP_QUERY
    # by taking all before first dot and appending value of
    # $_SATURNIN_CONF_SUFFIX (".ini" by default).  If also INIGREP_QUERY is
    # empty, all possible files are discovered and loaded.
    #
    # Following calls are equivalent
    #
    #     saturnin__conf foo.bar.baz
    #     saturnin__conf foo.bar.baz foo.ini
    #
    # and both result in reading of key *baz* from section *foo.bar* in file
    # *foo.ini*, which is selected from *_SATURNIN_CONF_PATH*.  Should there
    # be more foo.ini's, the first is selected.  Using `-j` switch
    #
    #     saturnin__conf -j foo.bar.baz
    #
    # would cause all foo.ini's on *_SATURNIN_CONF_PATH* be concatenated
    # instead.
    #
    local ig_mode               # inigrep retrieval mode
    local ig_query              # inigrep query (keypath or section name)
    local ig_limit              # inigrep line limiting argument
    local files=()              # file specification
    local Strategy=first        # merge strategy
    #shellcheck disable=SC2209
    while true; do case $1:$2 in
        "":*)   break ;;
        -j:*)   Strategy=join;           shift 1        ;;
        -1:*)   ig_limit=$1;             shift 1        ;;
        -e:*.*) ig_mode=$1; ig_query=$2; shift 2; break ;;
        -r:*.*) ig_mode=$1; ig_query=$2; shift 2; break ;;
        -K:*)   ig_mode=$1; ig_query=$2; shift 2; break ;;
        -S:*)   ig_mode=$1; ig_query=""; shift 1; break ;;
        -P:*)   ig_mode=$1; ig_query=""; shift 1; break ;;
        .*:*)   _saturnin__conf_usage -w "bad syntax: $*" ;;
        *.*:*)  ig_mode=-e; ig_query=$1; shift 1; break ;;
        --help:*) _saturnin__conf_usage -e 0            ;;
        *)      _saturnin__conf_usage -w "bad syntax: $*" ;;
    esac done
    test -n "$ig_mode" || _saturnin__conf_usage -w "could not determine inigrep mode"
    debug -v ig_limit ig_query ig_mode Strategy
    if test -n "$*"; then
        files=("$@")
    elif test -n "$ig_query"; then
        files=("${ig_query%%.*}$_SATURNIN_CONF_SUFFIX")
    else
        mapfile -t files <<<"$(saturnin__conf_files)"
        test ${#files[@]} -gt 0 || warn "no conf files found"
    fi
    debug -v files
    #shellcheck disable=SC2086
    _saturnin__conf__load "${files[@]}" | inigrep $ig_limit $ig_mode "$ig_query"
    return "${PIPESTATUS[0]}"
}

saturnin__conf_files() {
    #
    # Print all conf files found at $_SATURNIN_CONF_PATH
    #
    # Print list of filenames usable as FNAME argument to saturnin__conf().
    #
    local trydir        # each item of $_SATURNIN_CONF_PATH
    local trydirs=()    # possible dirs to try
    debug -v _SATURNIN_CONF_PATH
    for trydir in ${_SATURNIN_CONF_PATH//:/ }; do
        test -d "$trydir" && trydirs+=("$trydir")
    done
    test ${#trydirs[@]} -gt 0 \
     || return 1
    find -L "${trydirs[@]}" \
        -maxdepth 1 \
        -mindepth 1 \
        -name "*$_SATURNIN_CONF_SUFFIX" \
        -printf '%P\n' \
      | awk '!seen[$0]++' \
      | grep .
}

saturnin__conf_find() {
    #
    # Find all existing instances of sub-path $1 on $_SATURNIN_CONF_PATH
    #
    # Usage:
    #
    #     saturnin__conf_find SUBPATH
    #
    # Go through all elements of $_SATURNIN_CONF_PATH, looking for file or
    # directory, whose path is formed by joining SUBPATH to element of
    # $_SATURNIN_CONF_PATH.  Print each existing path, ignore rest.
    #
    # For example, with following setup:
    #
    #     _SATURNIN_CONF_PATH=foo:bar:baz
    #     mkdir -p foo/one bar/one
    #     mkdir -p bar/two/slashes
    #
    # call
    #
    #     saturnin__conf_find one
    #
    # would print `foo/one` and `bar/one`, while
    #
    #     saturnin__conf_find two/slashes
    #
    # would print `bar/two/slashes`.
    #
    # If at least one path was found, return zero.  Otherwise, return one,
    # or more in case of error.
    #
    local file=$1       # sub-path to find
    local trydir        # each item of $_SATURNIN_CONF_PATH
    local trypath       # each combined path
    debug -v _SATURNIN_CONF_PATH
    echos "$_SATURNIN_CONF_PATH" \
      | tr ':' '\n' \
      | while read -r trydir; do
            test -n "$trydir" || continue
            trypath="$trydir/$file"
            test -e "$trypath" || continue
            echos "$trypath"
        done \
      | grep .
}

saturnin__conf_mkpath() {
    #
    # Compose new value for $_SATURNIN_CONF_PATH from locations $@
    #
    # Usage:
    #
    #     saturnin__conf_mkpath DIR [DIR..]
    #
    # Go through each DIR and print it, unless it ends with "/ini.d",
    # in which case list its subdirectories, sorted by C locale (this allows
    # for modular configuration).
    #
    # Non-existent or non-directory items are silently ignored.
    #
    local location      # one location argument
    local path          # one path listed
    for location in "$@"; do
        test -d "$location" || continue
        case "$location" in
            */ini.d)    # modular location--sort subfolders
                find -L "$location" -mindepth 1 -maxdepth 1 -type d \
                  | LC_ALL=C sort
                ;;
            *)
                echo "$location"
                ;;
        esac
    done \
      | _saturnin__nl2colon
}

saturnin__get() {
    #
    # Show Saturnin internal info by key $1
    #
    # Key $1 can be whole `--saturnin-get-stuff` argument or just the part
    # after `--saturnin-get-`.
    #
    # This is aimed to help debugging and testing the app (or Saturnin
    # itself) by showing packaging and deployment related info.
    #
    local key=${1#--saturnin-get-}      # internal info key
    case "$key" in
        app-url)        echo "$_SATURNIN_APP_URL"           ;;
        app-version)    echo "$_SATURNIN_APP_VERSION"       ;;
        cache-home)     echo "$_SATURNIN_CACHE_HOME"        ;;
        conf-path)      echo "$_SATURNIN_CONF_PATH"         ;;
        config-home)    echo "$_SATURNIN_CONFIG_HOME"       ;;
        data-home)      echo "$_SATURNIN_DATA_HOME"         ;;
        saturnin-version)   echo "0.5.14"    ;;
        subcommand)     echo "$_SATURNIN_SUBCOMMAND"        ;;
        subcommands)    _saturnin__lssc                     ;;
        *)              warn "unknown devel key: $key"
                        return "$EXIT_USAGE" ;;
    esac
    return "$EXIT_OK"
}

saturnin__main() {
    #
    # Main meta-command entry function
    #
    # This is where almost every project and build specific detail
    # id set.  See saturnin-demo for example on how to call this
    # function properly.
    #
    local AppCodename       # Application codename
    local AppGitHash        # Git commit hash of application source tree
    local AppTagline        # Application tagline (short motto)
    local AppVersion        # Application version in SemVer format
    local Builtins=conf     # Colon-separated list of builtins to enable
    local CacheHome         # Path to user cache
    local ConfPath          # saturnin__conf() lookup path list (colon-separated)
    local ConfSuffix=.ini   # saturnin__conf() filename suffix
    local ConfigHome        # Path to user config
    local DataHome          # Path to user local data
    local HelpTopicPath     # saturnin_help topic lookup path
    local HelpTopicSuffix=.md   # saturnin_help topic lookup suffix
    local Libexec           # subcommand lookup path
    local LibexecPrefix     # subcommand lookup prefix
    local MetaHelp          # Meta-command help mode
    local ScCompgenPath     # Path to subcommand comgen scriptlets
    local Verbose=false     # default verbosity (`true` or `false`)
    local MsgMode=color     # default messaging mode (pretty/neaty module name)
    test -n "$PRETTY" && MsgMode=$PRETTY
    while true; do case $1 in
        --saturnin-app-codename)     AppCodename=$2;     shift 2 || return 2 ;;
        --saturnin-app-git-hash)     AppGitHash=$2;      shift 2 || return 2 ;;
        --saturnin-app-tagline)      AppTagline=$2;      shift 2 || return 2 ;;
        --saturnin-app-url)          AppUrl=$2;          shift 2 || return 2 ;;
        --saturnin-app-version)      AppVersion=$2;      shift 2 || return 2 ;;
        --saturnin-builtins)         Builtins=$2;        shift 2 || return 2 ;;
        --saturnin-cache-home)       CacheHome=$2;       shift 2 || return 2 ;;
        --saturnin-sccompgen-path)   ScCompgenPath=$2;   shift 2 || return 2 ;;
        --saturnin-conf-path)        ConfPath=$2;        shift 2 || return 2 ;;
        --saturnin-conf-suffix)      ConfSuffix=$2;      shift 2 || return 2 ;;
        --saturnin-config-home)      ConfigHome=$2;      shift 2 || return 2 ;;
        --saturnin-data-home)        DataHome=$2;        shift 2 || return 2 ;;
        --saturnin-help-topicpath)   HelpTopicPath=$2;   shift 2 || return 2 ;;
        --saturnin-help-topicsuffix) HelpTopicSuffix=$2; shift 2 || return 2 ;;
        --saturnin-libexec)          Libexec=$2;         shift 2 || return 2 ;;
        --saturnin-libexec-prefix)   LibexecPrefix=$2;   shift 2 || return 2 ;;
        --saturnin-meta-help)        MetaHelp=$2;        shift 2 || return 2 ;;
        --saturnin-msgmode)          MsgMode=${2:-$MsgMode}; shift 2 || return 2 ;;
        --saturnin-verbose)          Verbose=${2:-$Verbose}; shift 2 || return 2 ;;
        *) break ;;
    esac done
    export PRETTY=$MsgMode
    export NEATY=$MsgMode
    shellfu reload pretty
    test -n "$AppVersion"    || die "--saturnin-app-version is not set"
    test -n "$CacheHome"     || die "--saturnin-cache-home is not set"
    test -n "$Libexec"       || die "--saturnin-libexec is not set"
    test -n "$LibexecPrefix" || die "--saturnin-libexec-prefix is not set"
    local _SATURNIN_APP_GIT_HASH=$AppGitHash
    local _SATURNIN_APP_URL=$AppUrl
    local _SATURNIN_APP_VERSION=$AppVersion
    local _SATURNIN_CACHE_HOME=$CacheHome
    local _SATURNIN_CONFIG_HOME=$ConfigHome
    local _SATURNIN_CONF_PATH=$ConfPath
    local _SATURNIN_CONF_SUFFIX=$ConfSuffix
    local _SATURNIN_DATA_HOME=$DataHome
    export PRETTY_VERBOSE=$Verbose
    export NEATY_VERBOSE=$Verbose
    while true; do case $1 in
        --saturnin-get-*) saturnin__get "$1";       exit  ;;
        --saturnin-sccompgen) shift; _saturnin__sccompgen "$@"; exit ;;
        --version)      _saturnin__version;         exit  ;;
        -C|--no-color)  export PRETTY=plain; shellfu reload pretty
                          export NEATY=plain; shift ;;
        -D|--full-debug) export PRETTY_DEBUG=true
                        export PRETTY_DEBUG_EXCLUDE=""
                        export NEATY_DEBUG=true
                        export NEATY_DEBUG_EXCLUDE=""
                                                    shift ;;
        -V|--version-semver) echo "$AppVersion";    exit  ;;
        -c|--color)     export PRETTY=forcecolor; shellfu reload pretty
                        export NEATY=forcecolor; shift ;;
        -d|--debug)     export PRETTY_DEBUG=true
                        export NEATY_DEBUG=true;   shift ;;
        -q|--quiet)     export PRETTY_VERBOSE=false
                        export NEATY_VERBOSE=false; shift ;;
        -v|--verbose)   export PRETTY_VERBOSE=true
                        export NEATY_VERBOSE=true; shift ;;
        -h|--help)      _saturnin__help;            exit  ;;
        --)             shift; break                ;;
        -*)             _saturnin__usage -w "unknown argument: $1" ;;
        "")             _saturnin__usage -w "too few arguments" ;;
        *)              break;                      ;;
    esac done
    PRETTY_DEBUG_EXCLUDE="" debug -l \
        "starting ${0##*/} v$_SATURNIN_APP_VERSION bash v$BASH_VERSION" \
        "*=$*"
    _saturnin__runsc "$@"
}

saturnin__runhook() {
    #
    # Run custom hook named $1 from respective configuration section
    #
    # Will load joined multi-line key "hook.$_SATURNIN_SUBCOMMAND.$1" and
    # unless syntax check fails, execute it as Bash code (in separate
    # process).
    #
    local name="$1"     # hook name
    local code          # ... code
    test -n "$_SATURNIN_SUBCOMMAND" || {
        warn "unknown subcommand, ignoring hook: $name"
        return 0
    }
    code="$(saturnin__conf -j "hook.$_SATURNIN_SUBCOMMAND.$name")"
    debug -v _SATURNIN_SUBCOMMAND code name
    bash -n <<<"$code" || {
        warn "syntax errors, ignoring hook: $name"
        return 0
    }
    bash <<<"$code"
}

saturnin__wraphook() {
    #
    # Wrap command $@ in hooks 'pre' and 'post'
    #
    # Run pre hook, then command $@, then post hook.  Always exit with
    # status of the payload command, even if hooks fail.  Ignore post-hook
    # if payload command failed.
    #
    local es=0      # exit status of this function
    saturnin__runhook pre
    "$@" || return $?
    es=$?
    saturnin__runhook post
    return $es
}


#          #                           that what you see below this line #
# INTERNAL # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#          #                        use in your code to anger the divine #

#
# Git commit hash of application source tree
#
_SATURNIN_APP_GIT_HASH=${_SATURNIN_APP_GIT_HASH:-}

#
# Your application URL
#
_SATURNIN_APP_URL=${_SATURNIN_APP_URL:-}

#
# Your application version
#
_SATURNIN_APP_VERSION=${_SATURNIN_APP_VERSION:-}

#
# Path to user cache
#
_SATURNIN_CACHE_HOME=${_SATURNIN_CACHE_HOME:-}

#
# Path to user config root
#
_SATURNIN_CONFIG_HOME=${_SATURNIN_CONFIG_HOME:-}

#
# Path to user local data root
#
_SATURNIN_DATA_HOME=${_SATURNIN_DATA_HOME:-}

#
# Path where saturnin__conf should look for files
#
# If filename does not contain slash, it is looked up in each (or all,
# based on strategy--see saturnin__conf() doc) path in this list. The
# list is colon-separated and non-dirs as well as empty strings are
# silently ignored.
#
_SATURNIN_CONF_PATH="${_SATURNIN_CONF_PATH:-}"

#
# Expected config filename extension (for guessing from path head)
#
# If no filename to read is given, saturnin__conf() will guess filename
# as the path head plus this suffix (e.g. `foo.ini` for `saturnin__conf
# foo.bar.baz`)
#
_SATURNIN_CONF_SUFFIX="${_SATURNIN_CONF_SUFFIX:-.ini}"

#
# Current running subcommand
#
# The value is set and exported into sub-command binary at moment of
# saturnin__main() call.
#
_SATURNIN_SUBCOMMAND="${_SATURNIN_SUBCOMMAND:-}"

_saturnin__cat_helpfile() {
    #
    # Print helpfile $1
    #
    local helpfile=$1   # path to help file
    cat "$helpfile" >&2 && return 0
    saturnin__bug "cannot print help file: $helpfile"
    return 3
}

_saturnin__conf__compgen() {
    #
    # Handle saturnin__conf's completion
    #
    local pos=$1
    local previous=$2
    case $pos:$previous in
        0:conf)     echo "-j -1 -e -r -K -P -S" ;;
        1:-j|1:-1)  echo "-e -r -K -P -S" ;;
        *:-e|*:-r)  saturnin__conf -j -P ;;
        *:-K)       saturnin__conf -j -S ;;
        *:-P|*:-S)  saturnin__conf_files ;;
    esac
}

_saturnin__conf__load() {
    #
    # Print contents of files specified in $@
    #
    # Each argument means possible file candidate.  If candidate contains
    # slash, it's treated as file path and is printed directly.  If it's
    # single dash, standard input is copied.
    #
    # In all other cases, filename is searched in all elements of variable
    # _SATURNIN_CONF_PATH; output then depends on chosen $Strategy: with
    # 'first' strategy, first existing file is printed, with 'join'
    # strategy, all existing files are printed.
    #
    local arg       # each passed argument
    local es=0      # exit status of this function
    for arg in "$@"; do
        case $arg in
            -|*/*)      # stdin, or path (with slash)
                cat "$arg" || es=3
                ;;
            *)          # name given, find all its incarnations
                saturnin__conf_find "$arg" \
                  | _saturnin__conf__merge; es=$?
                ;;
        esac
    done
    return $es
}

_saturnin__conf__merge() {
    #
    # Take paths and applying merge strategy, load file(s)
    #
    local path          # every path
    local found=false   # 'true' if we got any path
    while read -r path; do
        found=true
        case $Strategy in
            first)
                debug "winner: $path"
                cat "$path"
                cat >/dev/null  # throw away rest of paths
                ;;
            join)
                echo "# file: ${path/$HOME/~}"
                cat "$path" 2>/dev/null
                ;;
        esac
    done
    $found
}

_saturnin__conf_usage() {
    #
    # Show usage message, passing $@ to mkusage() and exit
    #
    PRETTY_USAGE="self=${0##*/} conf" \
    mkusage "$@"                                                              \
            "[options] [-e] SECTION.KEY [FNAME]"                              \
            "[options] -r SECTION.KEY [FNAME]"                                \
            "[options] -K SECTION [FNAME]"                                    \
            "[options] -P [FNAME]"                                            \
            "[options] -S [FNAME]"                                            \
        --                                                                    \
            "Use inigrep to query config files."                              \
        -o                                                                    \
            "-j     join all files before applying query"                     \
            "-1     ensure single line is returned"                           \
        -c                                                                    \
            "-e     use normal mode (default)"                                \
            "-r     use raw mode (preserves RHS whitespace and some comments)"\
            "-K     list available keys in SECTION"                           \
            "-S     list available sections in FNAME"                         \
            "-P     list available keypaths (SECTION.KEY) in FNAME"           \
        --                                                                    \
            "FNAME is filename, which is then searched on all paths specified"\
            "in _SATURNIN_CONF_PATH and depending on -j parameter, first one"  \
            "wins or all are joined.  If FNAME contains slash, this search is"\
            "not done and FNAME is taken as path to file that is then"        \
            "queried."                                                        \
            ""                                                                \
            "If FNAME is omitted, it is inferred from SECTION (e.g. 'foo.ini'"\
            "if 'foo.bar' was section name; note that section name itself may"\
            "contain dot)."
}

_saturnin__help() {
    #
    # Print meta-command help text based on $MetaHelp
    #
    # $MetaHelp variable is set by an argument to saturnin__main(), and
    # controls what is displayed when user calls meta-command with
    # --help argument.  Following formats are supported:
    #
    #     +HELPFILE
    #     =HELPFILE
    #
    # In both cases, HELPFILE must be absolute path to a file containing
    # human-readable description of the meta-command, which will be directly
    # presented to user.
    #
    # If the leading character is '=' (equal sign), the help text consists
    # of, and only of the HELPFILE.  If the character is '+' (plus sign),
    # the help text is pre-pended with auto-generated usage message.
    #
    # If the value is empty, only the auto-generated usage message
    # is printed.
    #
    local introline     # introduction line
    introline=$(basename "$0")
    test -n "$AppTagline" \
     && introline+=" - $AppTagline"
    case "$MetaHelp" in
        "")
            echo "$introline"$'\n' >&2
            _saturnin__usage -E -e 0
            ;;
        +/*)
            echo "$introline"$'\n' >&2
            _saturnin__usage -E
            echo >&2
            _saturnin__cat_helpfile "${MetaHelp:1}"
            ;;
        =/*)
            _saturnin__cat_helpfile "${MetaHelp:1}"
            ;;
        *)
            echo "$introline"$'\n' >&2
            _saturnin__usage -E
            saturnin__bug "malformed meta-help spec: $MetaHelp"
            return 3
            ;;
    esac
}

_saturnin__isa_builtin() {
    #
    # True if $1 is a builtin
    #
    local sc=$1
    test "$sc" == conf && return 0
    test "$sc" == help && return 0
    return 1
}

_saturnin__isen_builtin() {
    #
    # True if $1 is an enabled builtin
    #
    local sc=$1
    local enbltns=":$Builtins:"
    debug -v enbltns
    _saturnin__isa_builtin "$sc"            || return 1
    test "${enbltns//:$sc:/}" == "$enbltns" && return 1
    return 0
}

_saturnin__lssc() {
    #
    # List subcommands
    #
    {
        find "$Libexec" \
            -mindepth 1 \
            -maxdepth 1 \
            -executable \
            -name "$LibexecPrefix*" \
          | sed -e "s|^.*/||; s|^$LibexecPrefix||"
        echo "$Builtins" \
          | sed 's/ *//; s/:/\n/' \
          | grep .
    } \
      | sort
}

_saturnin__nl2colon() {
    #
    # Convert newline-based list of paths to colon:based:list
    #
    # Empty paths must not be included in the resulting list, so we need to
    # drop them and also get the colons right.
    #
    local idx=0     # current item index (zero-based)
    local path      # each path on stdin
    while read -r path; do
        test -z "$path" && continue
        test $idx -gt 0 && echo -n ':'
        echo -n "$path"
        ((idx++))
    done
}

_saturnin__runsc() {
    #
    # Run subcommand $1 with arguments $2..
    #
    local subcommand="$1"; shift    # subcommand to run
    local binpath                   # path to subcommand's binary
    local isenblt=F                 # 'T' if $subcommand is enabled builtin
    local isenbin=F                 # 'T' if $subcommand is enabled binary
    binpath+="$Libexec/"
    binpath+="$LibexecPrefix$subcommand"
    debug -v binpath
    debug "\$*='$*'"
    _saturnin__isen_builtin "$subcommand" && isenblt=T
    test -x "$binpath"                    && isenbin=T
    debug -v isenbin isenblt
    case $isenbin:$isenblt in
        T:T)    warn "builtin enabled but overrided: $subcommand" ;;
        F:F)    _saturnin__usage -w "invalid sub-command: $subcommand" ;;
    esac
    case $isenblt:$subcommand in
        T:conf)     saturnin__conf "$@";                         return    ;;
        T:help)     _saturnin_help__main  \
                        --topicpath "$HelpTopicPath"  \
                        --topicsuffix "$HelpTopicSuffix"  \
                        "$@";                                    return    ;;
        T:*)        saturnin__bug "invalid built-in called: $subcommand"   ;;
        F:*)        PRETTY_DEBUG_EXCLUDE="$PRETTY_DEBUG_EXCLUDE,sfpi,inigrep,saturnin" \
                    PRETTY_USAGE="subcommand" \
                    _SATURNIN_APP_GIT_HASH=$_SATURNIN_APP_GIT_HASH \
                    _SATURNIN_APP_URL=$_SATURNIN_APP_URL \
                    _SATURNIN_APP_VERSION=$_SATURNIN_APP_VERSION \
                    _SATURNIN_CACHE_HOME=$_SATURNIN_CACHE_HOME \
                    _SATURNIN_CONFIG_HOME=$_SATURNIN_CONFIG_HOME \
                    _SATURNIN_CONF_PATH=$_SATURNIN_CONF_PATH \
                    _SATURNIN_CONF_SUFFIX=$_SATURNIN_CONF_SUFFIX \
                    _SATURNIN_DATA_HOME=$_SATURNIN_DATA_HOME \
                    _SATURNIN_SUBCOMMAND=$subcommand \
                        "$binpath" "$@"                                     ;;
        *)          saturnin__bug "invalid internals: $isenblt:$subcommand" ;;
    esac
}

_saturnin__sccompgen() {
    #
    # Generate sub-command completion options
    #
    local mc=$1;       shift
    local partial=$1;  shift
    local previous=$1; shift
    local sc=$1;       shift
    local words=("$@")
    local pos=0
    local word
    local script="$ScCompgenPath/$sc.sh"
    local cgargs=()
    for word in "$@"; do
        test "$word" == "$sc" && break
        ((pos++))
    done
    test -n "$partial" && (( pos-- ))
    cgargs=("$pos" "$previous" "$partial" "${words[@]}")
    if _saturnin__isen_builtin "$sc"; then
        case $sc in
            conf) _saturnin__conf__compgen "${cgargs[@]}" ;;
            help) _saturnin_help__compgen \
                    "$HelpTopicPath" \
                    "$HelpTopicSuffix" "${cgargs[@]}" ;;
        esac
        return $?
    fi
    test -f "$script" || {
        compgen -f "$partial"
        return 0
    }
    #shellcheck disable=SC2034,SC1090
    (
        export PRETTY_DEBUG=false
        export PRETTY_VERBOSE=false
        export NEATY_DEBUG=false
        export NEATY_VERBOSE=false
        export PRETTY=plain
        _SATURNIN_SCCOMPGEN__MC=$mc
        _SATURNIN_SCCOMPGEN__PARTIAL=$partial
        _SATURNIN_SCCOMPGEN__POS=$pos
        _SATURNIN_SCCOMPGEN__PREVIOUS=$previous
        _SATURNIN_SCCOMPGEN__SC=$sc
        . "$script" \
         || saturnin__bug "completion scriptlet did not exit wit zero status"
        type -t _saturnin__sccompgen_handle >/dev/null \
         || saturnin__bug "completion scriptlet did not define handler: _saturnin__sccompgen_handle()"
        _saturnin__sccompgen_handle "${cgargs[@]}"
    )
}

_saturnin__usage() {
    #
    # Show usage message and exit
    #
    #shellcheck disable=SC2034
    PRETTY_USAGE="self=${0##*/}"
    #shellcheck disable=SC2046
    mkusage "$@"                                            \
            "[options] COMMAND [ARG...]"                    \
        -o                                                  \
            "-C, --no-color   disable ANSI colors"          \
            "-D, --full-debug   turn on gory debugging"     \
            "-V, --version  show version and exit"          \
            "-c, --color    force ANSI colors"              \
            "-d, --debug    turn on debugging"              \
            "-h, --help     show this help message and exit"\
            "-q, --quiet    turn off verbosity"             \
            "-v, --verbose  turn on verbosity"              \
        -c                                                  \
            $(_saturnin__lssc | grep -v ^_)
}

_saturnin__version() {
    #
    # Print human-readable version info
    #
    # Basic version info is already stored in $AppVersion,
    # this function prints more descriptive paragraph including Saturnin's
    # own version.
    #
    echo -n "$(basename "$0")"
    test -n "$AppTagline" \
     && echo -n " ($AppTagline)"
    echo -n " $AppVersion"
    test -n "$AppCodename" \
     && echo -n " - $AppCodename"
    echo
    echo -n "Powered by Saturnin (Spirit of your toolbox)"
    echo -n " 0.5.14"
    echo -n " - DoktorV"
    echo
    return "$EXIT_OK"
}

#shellfu module-version=0.5.14
